@@ -30,10 +30,10 @@
 
 G_BEGIN_DECLS
 
-GST_EXPORT
+GST_RTSP_SERVER_API
 GstRTSPResult    gst_rtsp_params_set      (GstRTSPClient * client, GstRTSPContext * ctx);
 
-GST_EXPORT
+GST_RTSP_SERVER_API
 GstRTSPResult    gst_rtsp_params_get      (GstRTSPClient * client, GstRTSPContext * ctx);
 
 G_END_DECLS
